@@ -1,3 +1,4 @@
+"""The Aviary class, the core of how PyFlyt handles UAVs in the PyBullet simulation environment."""
 from __future__ import annotations
 
 import time
@@ -7,13 +8,15 @@ import pybullet as p
 import pybullet_data
 from pybullet_utils import bullet_client
 
-from .abstractions import DroneClass
+from .abstractions.base_drone import DroneClass
 from .drones.fixedwing import FixedWing
-from .drones.quadplane import Quadplane
 from .drones.quadx import QuadX
+from .drones.rocket import Rocket
 
 
 class Aviary(bullet_client.BulletClient):
+    """Aviary class, the core of how PyFlyt handles UAVs in the PyBullet simulation environment."""
+
     def __init__(
         self,
         start_pos: np.ndarray,
@@ -31,7 +34,31 @@ class Aviary(bullet_client.BulletClient):
         camera_resolution: tuple[int, int] = (128, 128),
         worldScale: float = 1.0,
         seed: None | int = None,
+        drone_options: dict = {},
     ):
+        """Initializes a PyBullet environment that hosts UAVs and other entities.
+
+        The Aviary class itself inherits from a BulletClient, so any function that a PyBullet client has, this class will have.
+        The Aviary also handles dealing with physics and control looprates, as well as automatic construction of several default UAVs and their corresponding cameras.
+
+        Args:
+            start_pos (np.ndarray): start_pos
+            start_orn (np.ndarray): start_orn
+            render (bool): render
+            physics_hz (int): physics_hz
+            ctrl_hz (int): ctrl_hz
+            drone_type (str): drone_type
+            drone_model (str): drone_model
+            model_dir (None | str): model_dir
+            use_camera (bool): use_camera
+            use_gimbal (bool): use_gimbal
+            camera_angle_degrees (int): camera_angle_degrees
+            camera_FOV_degrees (int): camera_FOV_degrees
+            camera_resolution (tuple[int, int]): camera_resolution
+            worldScale (float): worldScale
+            seed (None | int): seed
+            drone_options (dict): drone_options
+        """
         super().__init__(p.GUI if render else p.DIRECT)
         print("\033[A                             \033[A")
 
@@ -49,10 +76,10 @@ class Aviary(bullet_client.BulletClient):
         # define the drone types
         if drone_type == "quadx":
             self.drone_constructor = QuadX
-        elif drone_type == "quadplane":
-            self.drone_constructor = Quadplane
         elif drone_type == "fixedwing":
             self.drone_constructor = FixedWing
+        elif drone_type == "rocket":
+            self.drone_constructor = Rocket
 
         # default physics looprate is 240 Hz
         # do not change because pybullet doesn't like it
@@ -63,6 +90,7 @@ class Aviary(bullet_client.BulletClient):
         self.ctrl_update_ratio = int(physics_hz / ctrl_hz)
         self.now = time.time()
 
+        # pybullet stuff
         self.start_pos = start_pos
         self.start_orn = start_orn
         self.use_camera = use_camera
@@ -72,10 +100,13 @@ class Aviary(bullet_client.BulletClient):
         self.camera_frame_size = camera_resolution
         self.worldScale = worldScale
 
+        # directories and paths
         self.model_dir = model_dir
         self.drone_model = drone_model
+        self.drone_options = drone_options
         self.setAdditionalSearchPath(pybullet_data.getDataPath())
 
+        # render
         self.render = render
         self.rtf_debug_line = self.addUserDebugText(
             text="RTF here", textPosition=[0, 0, 0], textColorRGB=[1, 0, 0]
@@ -84,6 +115,11 @@ class Aviary(bullet_client.BulletClient):
         self.reset(seed)
 
     def reset(self, seed: None | int = None):
+        """Resets the simulation.
+
+        Args:
+            seed (None | int): seed
+        """
         self.resetSimulation()
         self.setGravity(0, 0, -9.81)
         self.steps = 0
@@ -116,6 +152,7 @@ class Aviary(bullet_client.BulletClient):
                     start_orn=start_orn,
                     ctrl_hz=self.ctrl_hz,
                     physics_hz=self.physics_hz,
+                    drone_model=self.drone_model,
                     model_dir=self.model_dir,
                     use_camera=self.use_camera,
                     use_gimbal=self.use_gimbal,
@@ -123,6 +160,7 @@ class Aviary(bullet_client.BulletClient):
                     camera_FOV_degrees=self.camera_FOV,
                     camera_resolution=self.camera_frame_size,
                     np_random=self.np_random,
+                    **self.drone_options,
                 )
             )
         # arm everything
@@ -130,6 +168,10 @@ class Aviary(bullet_client.BulletClient):
         self.set_armed(True)
 
     def register_all_new_bodies(self):
+        """Registers all new bodies in the environment to be able to handle collisions later.
+
+        Call this when there is an update in the number of bodies in the environment.
+        """
         # collision array
         self.collision_array = np.zeros(
             (self.getNumBodies(), self.getNumBodies()), dtype=bool
@@ -137,12 +179,19 @@ class Aviary(bullet_client.BulletClient):
 
     @property
     def num_drones(self) -> int:
+        """Returns the number of drones in the environment.
+
+        Returns:
+            int: number of drones in the environment
+        """
         return len(self.drones)
 
     @property
     def states(self) -> np.ndarray:
-        """
-        returns a list of states for each drone in the aviary
+        """Returns a list of states for all drones in the environment.
+
+        Returns:
+            np.ndarray: list of states
         """
         states = []
         for drone in self.drones:
@@ -152,9 +201,36 @@ class Aviary(bullet_client.BulletClient):
 
         return states
 
-    def set_armed(self, settings: int | bool | list[int | bool]):
+    @property
+    def aux_states(self) -> np.ndarray:
+        """Returns a list of auxiliary states for all drones in the environment.
+
+        Returns:
+            np.ndarray: list of auxiliary states
         """
-        sets the arming status for all the drones
+        aux_states = []
+        for drone in self.drones:
+            aux_states.append(drone.aux_state)
+
+        aux_states = np.stack(aux_states, axis=0)
+
+        return aux_states
+
+    def print_all_bodies(self):
+        """Debugging function used to print out all bodies in the environment along with their IDs."""
+        bodies = dict()
+        for i in range(self.getNumBodies()):
+            bodies[i] = self.getBodyInfo(i)[-1].decode("UTF-8")
+
+        from pprint import pprint
+
+        pprint(bodies)
+
+    def set_armed(self, settings: int | bool | list[int | bool]):
+        """Sets the arming state of each drone in the environment.
+
+        Args:
+            settings (int | bool | list[int | bool]): arm setting
         """
         if isinstance(settings, list):
             assert len(settings) == len(
@@ -167,8 +243,10 @@ class Aviary(bullet_client.BulletClient):
             self.armed_drones = [drone for drone in self.drones] if settings else []
 
     def set_mode(self, flight_modes: int | list[int]):
-        """
-        sets the flight mode for each drone
+        """Sets the flight mode of each drone in the environment.
+
+        Args:
+            flight_modes (int | list[int]): flight mode
         """
         if isinstance(flight_modes, list):
             assert len(flight_modes) == len(
@@ -181,16 +259,16 @@ class Aviary(bullet_client.BulletClient):
                 drone.set_mode(flight_modes)
 
     def set_setpoints(self, setpoints: np.ndarray):
-        """
-        commands each drone to go to a setpoint as specified in a list
+        """Sets the setpoints of each drone in the environment.
+
+        Args:
+            setpoints (np.ndarray): list of setpoints
         """
         for i, drone in enumerate(self.drones):
             drone.setpoint = setpoints[i]
 
     def step(self):
-        """
-        Steps the environment
-        """
+        """Steps the environment, this automatically handles physics and control looprates, one step is equivalent to one control loop step."""
         # compute rtf if we're rendering
         if self.render:
             elapsed = time.time() - self.now
