@@ -1,3 +1,4 @@
+"""Implementation of a CrazyFlie 2.x UAV."""
 from __future__ import annotations
 
 import math
@@ -6,8 +7,11 @@ import numpy as np
 import yaml
 from pybullet_utils import bullet_client
 
-from ..abstractions import CtrlClass, DroneClass
-from ..pid import PID
+from ..abstractions.base_controller import CtrlClass
+from ..abstractions.base_drone import DroneClass
+from ..abstractions.camera import Camera
+from ..abstractions.motors import Motors
+from ..abstractions.pid import PID
 
 
 class QuadX(DroneClass):
@@ -71,24 +75,43 @@ class QuadX(DroneClass):
             ctrl_params = all_params["control_params"]
 
             # motor thrust and torque constants
-            self.t2w = motor_params["thrust_to_weight"]
-            self.kf = motor_params["thrust_const"]
-            self.km = motor_params["torque_const"]
+            motor_ids = [0, 1, 2, 3]
+            thrust_coef = np.array([motor_params["thrust_coef"]] * 4)
+            torque_coef = np.array(
+                [
+                    +motor_params["torque_coef"],
+                    +motor_params["torque_coef"],
+                    -motor_params["torque_coef"],
+                    -motor_params["torque_coef"],
+                ]
+            )
+            thrust_unit = np.array(
+                [
+                    [0.0, 0.0, 1.0],
+                    [0.0, 0.0, 1.0],
+                    [0.0, 0.0, 1.0],
+                    [0.0, 0.0, 1.0],
+                ]
+            )
+            noise_ratio = np.array([1.0] * 4) * motor_params["noise_ratio"]
+            max_rpm = np.array([1.0] * 4) * np.sqrt(
+                (motor_params["total_thrust"]) / (4 * motor_params["thrust_coef"])
+            )
+            tau = np.array([1.0] * 4) * motor_params["tau"]
+            self.motors = Motors(
+                p=self.p,
+                physics_period=self.physics_period,
+                np_random=self.np_random,
+                uav_id=self.Id,
+                motor_ids=motor_ids,
+                tau=tau,
+                max_rpm=max_rpm,
+                thrust_coef=thrust_coef,
+                torque_coef=torque_coef,
+                thrust_unit=thrust_unit,
+                noise_ratio=noise_ratio,
+            )
 
-            # the joint IDs corresponding to motorID 1234
-            self.thr_coeff = np.array([[0.0, 0.0, 1.0]]) * self.kf
-            self.tor_coeff = np.array([[0.0, 0.0, 1.0]]) * self.km
-            self.tor_dir = np.array([[1.0], [1.0], [-1.0], [-1.0]])
-            self.noise_ratio = motor_params["motor_noise_ratio"]
-
-            # pseudo drag coef
-            self.drag_const_xyz = drag_params["drag_const_xyz"]
-            self.drag_const_pqr = drag_params["drag_const_pqr"]
-
-            # maximum motor RPM
-            self.max_rpm = np.sqrt((self.t2w * 9.81) / (4 * self.kf))
-            # motor modelled with first order ode, below is time const
-            self.motor_tau = 0.01
             # motor mapping from command to individual motors
             self.motor_map = np.array(
                 [
@@ -99,6 +122,10 @@ class QuadX(DroneClass):
                 ]
             )
 
+            # pseudo drag coef
+            self.drag_coef_xyz = drag_params["drag_coef_xyz"]
+            self.drag_coef_pqr = drag_params["drag_coef_pqr"]
+
             self.Kp_ang_vel = np.array(ctrl_params["ang_vel"]["kp"])
             self.Ki_ang_vel = np.array(ctrl_params["ang_vel"]["ki"])
             self.Kd_ang_vel = np.array(ctrl_params["ang_vel"]["kd"])
@@ -142,106 +169,34 @@ class QuadX(DroneClass):
         """ CAMERA """
         self.use_camera = use_camera
         if self.use_camera:
-            self.proj_mat = self.p.computeProjectionMatrixFOV(
-                fov=camera_FOV_degrees, aspect=1.0, nearVal=0.1, farVal=255.0
+            self.camera = Camera(
+                p=self.p,
+                uav_id=self.Id,
+                camera_id=0,
+                use_gimbal=use_gimbal,
+                camera_FOV_degrees=camera_FOV_degrees,
+                camera_angle_degrees=camera_angle_degrees,
+                camera_resolution=camera_resolution,
             )
-            self.use_gimbal = use_gimbal
-            self.camera_angle_degrees = camera_angle_degrees
-            self.camera_FOV_degrees = camera_FOV_degrees
-            self.camera_resolution = np.array(camera_resolution)
-
-        """ CUSTOM CONTROLLERS """
-        # dictionary mapping of controller_id to controller objects
-        self.registered_controllers = dict()
-        self.instanced_controllers = dict()
-        self.registered_base_modes = dict()
 
         self.reset()
 
     def reset(self):
-        """reset."""
+        """Resets the vehicle to the initial state."""
         self.set_mode(0)
-        self.state = np.zeros((4, 3))
-        self.setpoint = np.zeros((4))
-        self.rpm = np.zeros((4))
-        self.pwm = np.zeros((4))
+        self.setpoint = np.zeros(4)
+        self.pwm = np.zeros(4)
 
         self.p.resetBasePositionAndOrientation(self.Id, self.start_pos, self.start_orn)
+        self.disable_artificial_damping()
+        self.motors.reset()
         self.update_state()
 
         if self.use_camera:
-            self.capture_image()
+            self.rgbaImg, self.depthImg, self.segImg = self.camera.capture_image()
 
-    def update_drag(self):
-        """adds drag to the model, this is not physically correct but only approximation"""
-        lin_vel, ang_vel = self.p.getBaseVelocity(self.Id)
-        drag_xyz = -self.drag_const_xyz * (np.array(lin_vel) ** 2)
-        drag_pqr = -self.drag_const_pqr * (np.array(ang_vel) ** 2)
-
-        # warning, the physics is funky for bounces
-        if len(self.p.getContactPoints()) == 0:
-            self.p.applyExternalForce(
-                self.Id, -1, drag_xyz, [0.0, 0.0, 0.0], self.p.LINK_FRAME
-            )
-            self.p.applyExternalTorque(self.Id, -1, drag_pqr, self.p.LINK_FRAME)
-
-    def rpm2forces(self, rpm):
-        """maps rpm to individual motor forces and torques"""
-        rpm = np.expand_dims(rpm, axis=1)
-        thrust = (rpm**2) * self.thr_coeff
-        torque = (rpm**2) * self.tor_coeff * self.tor_dir
-
-        # add some random noise to the motor outputs
-        thrust += self.np_random.randn(*thrust.shape) * self.noise_ratio * thrust
-        torque += self.np_random.randn(*torque.shape) * self.noise_ratio * torque
-
-        for idx, (thr, tor) in enumerate(zip(thrust, torque)):
-            self.p.applyExternalForce(
-                self.Id, idx, thr, [0.0, 0.0, 0.0], self.p.LINK_FRAME
-            )
-            self.p.applyExternalTorque(self.Id, idx, tor, self.p.LINK_FRAME)
-
-    def pwm2rpm(self, pwm):
-        """model the motor using first order ODE, y' = T/tau * (setpoint - y)"""
-        self.rpm += (self.physics_hz / self.motor_tau) * (self.max_rpm * pwm - self.rpm)
-
-        return self.rpm
-
-    def cmd2pwm(self, cmd):
-        """maps angular torque commands to motor rpms"""
-        pwm = np.matmul(self.motor_map, cmd)
-
-        # deal with motor saturations
-        if (high := np.max(pwm)) > 1.0:
-            pwm /= high
-        if (low := np.min(pwm)) < 0.05:
-            pwm += (1.0 - pwm) / (1.0 - low) * (0.05 - low)
-
-        return pwm
-
-    def update_state(self):
-        """ang_vel, ang_pos, lin_vel, lin_pos"""
-        lin_pos, ang_pos = self.p.getBasePositionAndOrientation(self.Id)
-        lin_vel, ang_vel = self.p.getBaseVelocity(self.Id)
-
-        # express vels in local frame
-        rotation = np.array(self.p.getMatrixFromQuaternion(ang_pos)).reshape(3, 3).T
-        lin_vel = np.matmul(rotation, lin_vel)
-        ang_vel = np.matmul(rotation, ang_vel)
-
-        # ang_pos in euler form
-        ang_pos = self.p.getEulerFromQuaternion(ang_pos)
-
-        self.state = np.stack([ang_vel, ang_pos, lin_vel, lin_pos], axis=0)
-
-    def set_mode(self, mode):
-        """
-        vp, vq, vr = angular velocities
-        p, q, r = angular positions
-        u, v, w = local linear velocities
-        x, y, z = linear positions
-        vx, vy, vz = ground linear velocities
-        T = thrust
+    def set_mode(self, mode: int):
+        """Sets the current flight mode of the vehicle.
 
         sets the flight mode:
            -1 - m1, m2, m3, m4
@@ -253,6 +208,17 @@ class QuadX(DroneClass):
             5 - u, v, vr, vz
             6 - vx, vy, vr, vz
             7 - x, y, r, z
+
+        vp, vq, vr = angular velocities
+        p, q, r = angular positions
+        u, v, w = local linear velocities
+        x, y, z = linear positions
+        vx, vy, vz = ground linear velocities
+        T = thrust
+
+        Args:
+            mode (int): flight mode
+
         """
         if (mode < -1 or mode > 7) and mode not in self.registered_controllers.keys():
             raise ValueError(
@@ -370,8 +336,53 @@ class QuadX(DroneClass):
         for controller in self.PIDs:
             controller.reset()
 
+    def register_controller(
+        self,
+        controller_id: int,
+        controller_constructor: type[CtrlClass],
+        base_mode: int,
+    ):
+        """Registers a new controller for the UAV.
+
+        Args:
+            controller_id (int): controller_id
+            controller_constructor (type[CtrlClass]): controller_constructor
+            base_mode (int): base_mode
+        """
+        assert (
+            controller_id > 7
+        ), f"`controller_id` must be more than 7, currently {controller_id}"
+        assert (
+            base_mode >= -1 and base_mode <= 7
+        ), f"`base_mode` must be within -1 and 7, currently {base_mode}."
+
+        self.registered_controllers[controller_id] = controller_constructor
+        self.registered_base_modes[controller_id] = base_mode
+
+    def update_state(self):
+        """Updates the current state of the UAV.
+
+        This includes: ang_vel, ang_pos, lin_vel, lin_pos.
+        """
+        lin_pos, ang_pos = self.p.getBasePositionAndOrientation(self.Id)
+        lin_vel, ang_vel = self.p.getBaseVelocity(self.Id)
+
+        # express vels in local frame
+        rotation = np.array(self.p.getMatrixFromQuaternion(ang_pos)).reshape(3, 3).T
+        lin_vel = np.matmul(rotation, lin_vel)
+        ang_vel = np.matmul(rotation, ang_vel)
+
+        # ang_pos in euler form
+        ang_pos = self.p.getEulerFromQuaternion(ang_pos)
+
+        # create the state
+        self.state = np.stack([ang_vel, ang_pos, lin_vel, lin_pos], axis=0)
+
+        # update auxiliary information
+        self.aux_state = self.motors.get_states()
+
     def update_control(self):
-        """runs through controllers"""
+        """Runs through controllers."""
         # this is the thing we cascade down controllers
         a_output = self.setpoint[:3].copy()
         z_output = self.setpoint[-1].copy()
@@ -441,93 +452,37 @@ class QuadX(DroneClass):
             z_output = self.z_PIDs[0].step(self.state[2][-1], z_output)
             z_output = np.clip(z_output, 0, 1)
 
-        # mix the commands
-        self.pwm = self.cmd2pwm(np.array([*a_output, z_output]))
+        # mix the commands according to motor mix
+        cmd = np.array([*a_output, z_output])
+        self.pwm = np.matmul(self.motor_map, cmd)
+
+        # deal with motor saturations
+        if (high := np.max(self.pwm)) > 1.0:
+            self.pwm /= high
+        if (low := np.min(self.pwm)) < 0.05:
+            self.pwm += (1.0 - self.pwm) / (1.0 - low) * (0.05 - low)
+
+    def update_drag(self):
+        """Adds drag to the model, this is not physically correct but only approximation."""
+        drag_pqr = -self.drag_coef_pqr * (np.array(self.state[0]) ** 2)
+        drag_xyz = -self.drag_coef_xyz * (np.array(self.state[2]) ** 2)
+
+        # warning, the physics is funky for bounces
+        if len(self.p.getContactPoints()) == 0:
+            self.p.applyExternalForce(
+                self.Id, -1, drag_xyz, [0.0, 0.0, 0.0], self.p.LINK_FRAME
+            )
+            self.p.applyExternalTorque(self.Id, -1, drag_pqr, self.p.LINK_FRAME)
 
     def update_physics(self):
-        """update_physics."""
-        self.rpm2forces(self.pwm2rpm(self.pwm))
+        """Updates the physics of the vehicle."""
+        self.update_state()
+        self.motors.pwm2forces(self.pwm)
         self.update_drag()
 
-    @property
-    def view_mat(self):
-        """view_mat."""
-        # get the state of the camera on the robot
-        camera_state = self.p.getLinkState(self.Id, 0)
-
-        # pose and rot
-        position = camera_state[0]
-
-        # simulate gimballed camera if needed
-        up_vector = None
-        if self.use_gimbal:
-            # camera tilted downward for gimballed mode
-            rot = np.array(self.p.getEulerFromQuaternion(camera_state[1]))
-            rot[0] = 0.0
-            rot[1] = self.camera_angle_degrees / 180 * math.pi
-            rot = np.array(self.p.getQuaternionFromEuler(rot))
-            rot = np.array(self.p.getMatrixFromQuaternion(rot)).reshape(3, 3)
-
-            up_vector = np.matmul(rot, np.array([0.0, 0.0, 1.0]))
-        else:
-            # camera rotated upward for FPV mode
-            rot = np.array(self.p.getEulerFromQuaternion(camera_state[1]))
-            rot[1] += -self.camera_angle_degrees / 180 * math.pi
-            rot = np.array(self.p.getQuaternionFromEuler(rot))
-            rot = np.array(self.p.getMatrixFromQuaternion(rot)).reshape(3, 3)
-
-            up_vector = np.matmul(rot, np.array([0, 0, 1]))
-
-        # target position is 1000 units ahead of camera relative to the current camera pos
-        target = np.dot(rot, np.array([1000, 0, 0])) + np.array(position)
-
-        return self.p.computeViewMatrix(
-            cameraEyePosition=position,
-            cameraTargetPosition=target,
-            cameraUpVector=up_vector,
-        )
-
-    def capture_image(self):
-        """capture_image."""
-        _, _, self.rgbaImg, self.depthImg, self.segImg = self.p.getCameraImage(
-            width=self.camera_resolution[1],
-            height=self.camera_resolution[0],
-            viewMatrix=self.view_mat,
-            projectionMatrix=self.proj_mat,
-        )
-        self.rgbaImg = np.array(self.rgbaImg).reshape(*self.camera_resolution, -1)
-        self.depthImg = np.array(self.depthImg).reshape(*self.camera_resolution, -1)
-        self.segImg = np.array(self.segImg).reshape(*self.camera_resolution, -1)
-
-    def register_controller(
-        self,
-        controller_id: int,
-        controller_constructor: type[CtrlClass],
-        base_mode: int,
-    ):
-        """register_controller.
-
-        Args:
-            controller_id (int): controller_id
-            controller_constructor (type[CtrlClass]): controller_constructor
-            base_mode (int): base_mode
-        """
-        assert (
-            controller_id > 7
-        ), f"`controller_id` must be more than 7, currently {controller_id}"
-        assert (
-            base_mode >= -1 and base_mode <= 7
-        ), f"`base_mode` must be within -1 and 7, currently {base_mode}."
-
-        self.registered_controllers[controller_id] = controller_constructor
-        self.registered_base_modes[controller_id] = base_mode
-
     def update_avionics(self):
-        """
-        updates state and control
-        """
-        self.update_state()
+        """Updates state and control."""
         self.update_control()
 
         if self.use_camera:
-            self.capture_image()
+            self.rgbaImg, self.depthImg, self.segImg = self.camera.capture_image()
